__version__ = '0.11.4.dev2'

import os

from .document import Document
from .array import DocumentArray
from .dataclasses import dataclass, Image, Text, Audio, JSON, field


if 'DA_NO_RICH_HANDLER' not in os.environ:
    from rich.traceback import install

    install()



import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from docarray import Document


def image_getter(doc: 'Document', field_name: str):
    if 'image_type' in doc._metadata:
        if doc._metadata['image_type'] == 'uri':
            return doc._metadata['image_uri']
        elif doc._metadata['image_type'] == 'PIL':
            from PIL import Image

            return Image.fromarray(doc.tensor)
        elif doc._metadata['image_type'] == 'ndarray':
            return doc.tensor
    else:
        raise ValueError('Invalid image Document')


def text_getter(doc: 'Document', field_name: str):
    return doc.text


def audio_getter(doc: 'Document', field_name: str):
    from PIL import Image

    return Image.fromarray(doc.tensor)


def json_getter(doc: 'Document', field_name: str):
    if doc._metadata['json_type'] == 'str':
        return json.dumps(doc.tags[field_name])
    else:
        return doc.tags[field_name]



from ...helper import deprecate_by


class PlotMixin:
    """Provide helper functions for :class:`Document` to plot and visualize itself."""

    def _ipython_display_(self):
        """Displays the object in IPython as a side effect"""
        self.summary()

    def __rich_console__(self, console, options):

        yield f":page_facing_up: [b]Document[/b]: [cyan]{self.id}[cyan]"
        from rich.table import Table
        from rich import box

        my_table = Table(
            'Attribute', 'Value', width=80, box=box.ROUNDED, highlight=True
        )
        for f in self.non_empty_fields:
            if f in ('embedding', 'tensor'):
                from ...math.ndarray import to_numpy_array

                v = to_numpy_array(getattr(self, f))
                if v.squeeze().ndim == 1:
                    from .rich_embedding import ColorBoxEmbedding

                    v = ColorBoxEmbedding(v.squeeze())
                else:
                    v = str(getattr(self, f))
                my_table.add_row(f, v)
            elif f not in ('id', 'chunks', 'matches'):
                my_table.add_row(f, str(getattr(self, f)))
        if my_table.rows:
            yield my_table

    def summary(self) -> None:
        """Print non-empty fields and nested structure of this Document object."""
        from rich import print

        print(self._plot_recursion())

    def _plot_recursion(self, tree=None):
        if tree is None:
            from rich.tree import Tree

            tree = Tree(self)
        else:
            tree = tree.add(self)
        for a in ('matches', 'chunks'):
            if getattr(self, a):
                if a == 'chunks':
                    _icon = ':diamond_with_a_dot:'
                else:
                    _icon = ':large_orange_diamond:'
                _match_tree = tree.add(f'{_icon} [b]{a.capitalize()}[/b]')
                for d in getattr(self, a):
                    d._plot_recursion(_match_tree)
        return tree

    def display(self):
        """Plot image data from :attr:`.tensor` or :attr:`.uri`."""
        from IPython.display import Image, display

        if self.uri:
            if self.mime_type.startswith('audio') or self.uri.startswith('data:audio/'):
                uri = _convert_display_uri(self.uri, self.mime_type)
                _html5_audio_player(uri)
            elif self.mime_type.startswith('video') or self.uri.startswith(
                'data:video/'
            ):
                uri = _convert_display_uri(self.uri, self.mime_type)
                _html5_video_player(uri)
            elif self.uri.startswith('data:image/'):
                _html5_image(self.uri)
            else:
                display(Image(self.uri))
        elif self.tensor is not None:
            try:
                import PIL.Image

                p = PIL.Image.fromarray(self.tensor)
                if p.mode != 'RGB':
                    raise
                display(p)
            except:
                import matplotlib.pyplot as plt

                plt.matshow(self.tensor)
        else:
            self.summary()

    plot = deprecate_by(display, removed_at='0.5')


def _convert_display_uri(uri, mime_type):
    import urllib
    from .helper import _to_datauri, _uri_to_blob

    scheme = urllib.parse.urlparse(uri).scheme

    if scheme not in ['data', 'http', 'https']:
        blob = _uri_to_blob(uri)
        return _to_datauri(mime_type, blob)
    return uri


def _html5_image(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <image src="{uri}" height="200px">
    </body>
    '''
    display(HTML(src))  # noqa


def _html5_video_player(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <video width="320" height="240" autoplay muted controls>
    <source src="{uri}">
    Your browser does not support the video tag.
    </video>
    </body>
    '''
    display(HTML(src))  # noqa


def _html5_audio_player(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <audio controls="controls" style="width:320px" >
      <source src="{uri}"/>
      Your browser does not support the audio element.
    </audio>
    </body>
    '''
    display(HTML(src))  # noqa



import json
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from docarray import Document


def image_setter(field_name: str, value) -> 'Document':
    from PIL.Image import Image
    from docarray import Document

    doc = Document(modality='image')

    if isinstance(value, str):
        doc.uri = value
        doc._metadata['image_type'] = 'uri'
        doc._metadata['image_uri'] = value
        doc.load_uri_to_image_tensor()
    elif isinstance(value, Image):
        doc.tensor = np.array(value)
        doc._metadata['image_type'] = 'PIL'
    else:
        doc.tensor = value
        doc._metadata['image_type'] = 'ndarray'
    return doc


def text_setter(field_name: str, value) -> 'Document':
    from docarray import Document

    return Document(text=value, modality='text')


def audio_setter(field_name: str, value) -> 'Document':
    import librosa
    from docarray import Document

    audio, sr = librosa.load(value)
    return Document(
        tensor=audio,
        _metadata={'audio_sample_rate': sr, 'audio_uri': str(value)},
        modality='audio',
    )


def json_setter(field_name: str, value) -> 'Document':
    from docarray import Document

    doc = Document()
    if isinstance(value, str):
        value = json.loads(value)
        doc._metadata['json_type'] = 'str'
    else:
        doc._metadata['json_type'] = 'dict'
    doc.tags[field_name] = value
    return doc



from typing import (
    TYPE_CHECKING,
    TypeVar,
    Sequence,
    List,
)

import numpy as np

from .... import Document, DocumentArray
from ....math import ndarray
from ....math.helper import EPSILON
from ....math.ndarray import to_numpy_array
from ....score import NamedScore

if TYPE_CHECKING:
    import tensorflow
    import torch

    ElasticArrayType = TypeVar(
        'ElasticArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )


class FindMixin:
    def _find_similar_vectors(self, query: 'ElasticArrayType', limit=10):
        query = to_numpy_array(query)
        is_all_zero = np.all(query == 0)
        if is_all_zero:
            query = query + EPSILON

        resp = self._client.knn_search(
            index=self._config.index_name,
            knn={
                "field": "embedding",
                "query_vector": query,
                "k": limit,
                "num_candidates": 10000,
            },
        )
        list_of_hits = resp['hits']['hits']

        da = DocumentArray()
        for result in list_of_hits:
            doc = Document.from_base64(result['_source']['blob'])
            doc.scores['score'] = NamedScore(value=result['_score'])
            doc.embedding = result['_source']['embedding']
            da.append(doc)

        return da

    def _find(
        self, query: 'ElasticArrayType', limit: int = 10, **kwargs
    ) -> List['DocumentArray']:
        """Returns approximate nearest neighbors given a batch of input queries.
        :param query: input supported to be stored in Elastic. This includes any from the list '[np.ndarray, tensorflow.Tensor, torch.Tensor, Sequence[float]]'
        :param limit: number of retrieved items

        :return: DocumentArray containing the closest documents to the query if it is a single query, otherwise a list of DocumentArrays containing
           the closest Document objects for each of the queries in `query`.

        """

        num_rows, _ = ndarray.get_array_rows(query)

        if num_rows == 1:
            return [self._find_similar_vectors(query[0], limit=limit)]
        else:
            closest_docs = []
            for q in query:
                da = self._find_similar_vectors(q, limit=limit)
                closest_docs.append(da)
            return closest_docs
